/*
    dio v1.4
    Copyright 2001, 2002 Donald McIntosh
    Email: donmcintosh@hotmail.com
    Source code for dio: device I/O analysis tool

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <curses.h>
#include <kstat.h>
#include <time.h>
#include <signal.h>
#include <sys/param.h>
#include "dio.h"
#include "diskmap.h"

#define WIDTH 80
#define	HEIGHT 24
#define HDHEIGHT 4
#define TOTHEIGHT 3
#define SCRLHEIGHT 17

void initwindow(WINDOW *win);
void setheadwin(WINDOW *win, char* dev);
void updatescreen(WINDOW*, WINDOW*, kstat_ctl_t *kc, kstat_t *ksp, int ival);
void calcstats(char* line, kstat_io_t	new, kstat_io_t old, int ival);
void printtots(WINDOW *win);
void listio();
void usage();
void bomb(int exitv, char *msg, int showusage);
void close_end(int signum);
void check_input();

int main(int argv, char** argc)
{
	char 	*dev;
	char 	*intervalc;
	kstat_ctl_t	*kc;
	kstat_t		*ksp;
	int			interval;
	int			maxy, maxx;
	int			pad = 0; /*arg padding*/
	char	msg[200];
	struct sigaction quitaction;

	WINDOW		*mainwin = initscr();
	/* initialize all three windows */

	WINDOW		*headwin = newwin(HDHEIGHT,WIDTH,0,0);
	WINDOW		*scrollwin = newwin(SCRLHEIGHT,WIDTH,HDHEIGHT,0);
	WINDOW		*totwin = newwin(TOTHEIGHT,WIDTH,SCRLHEIGHT+HDHEIGHT,0);

	/* dont echo and don't block characters */
	noecho();
	cbreak();

	/* setup signal handlers */
	quitaction.sa_handler = close_end;
	sigemptyset(&quitaction.sa_mask);
	quitaction.sa_flags = 0;
	
	/* run close_end on receipt of SIGINT */
	sigaction(SIGINT, &quitaction, NULL);
	sigaction(SIGTERM, &quitaction, NULL);
	sigaction(SIGHUP, &quitaction, NULL);

	if(argv == 2){
		if(argc[1][0] == '-'){
			if(argc[1][1] == 'l'){
				listio();
			}
			else{
				bomb(1, "", 1);
			}
		}
	}

	if (argv == 5){
		pad = 2;

		if(argc[1][0] == '-'){
			if(argc[1][1] == 'o'){
				ofile = fopen(argc[2], "w");
				if(ofile == NULL){
					sprintf(msg,"Cannot open %s.\n",argc[2]);
					bomb(1, msg, 0);
				}
				fprintf(ofile, "# dio v%s data output file\n",version);
			}
			else{
				sprintf(msg,"Argument %c is not valid.\n",argc[1][1]);
				bomb(1, msg, 1);
			}
		}
	}
	else if( argv != 3 ){
		bomb(1, "", 1);
	}
	
	dev = argc[pad +1];
	if(dev == NULL){
		sprintf(msg, "You must specify a device name.\n");
		bomb(1, msg, 1);
	}
	intervalc = argc[pad +2];
	if(intervalc == NULL){
		sprintf(msg, "You must specify a time interval.\n");
		bomb(1, msg, 1);
	}
	interval = atoi(intervalc);
	if(interval < 1){
 		sprintf(msg, "You must specify a time interval greater than or equal to 1.\n");
		bomb(1, msg, 0);
	}
	
	kc = kstat_open();
	ksp = kstat_lookup(kc, NULL, -1, dev);

	if(ksp == NULL){
		sprintf(msg,"You must specify a valid device name.\n");
		bomb(1, msg, 0);
	}

	if(ksp->ks_type != KSTAT_TYPE_IO){
		sprintf(msg,"You must specify an IO device.\n");
		bomb(1, msg, 0);
	}

	/* confirm dimensions of window */
	getmaxyx(mainwin, maxy, maxx);
	if(maxy < HEIGHT){
		sprintf(msg, "Please extend the height of your terminal window.\nIt must be 24 characters in height to run dio.\n");
		bomb(1, msg, 0);
	}
	if(maxx < WIDTH){
		sprintf(msg, "Please extend the width of your terminal window.\nIt must be 80 characters wide to run dio.\n");
		bomb(1, msg, 0);
	}

	/* initialize all window*/
	initwindow(mainwin);
	initwindow(headwin);
	initwindow(scrollwin);
	initwindow(totwin);

	/* set scrolling active on scrollable window*/
	scrollok(scrollwin, TRUE);

	/* write the heading window */
	setheadwin(headwin, dev);

	updatescreen(scrollwin, totwin, kc, ksp, interval);

	endwin();
	if(ofile != NULL){
		fclose(ofile);
	}
	exit(0);
}

void initwindow(WINDOW* win)
{
	nodelay(win, TRUE);
	wclear(win);
	wrefresh(win);
}
	
void setheadwin(WINDOW *win, char* dev)
{
	char colhead[WIDTH+1];
	char header[WIDTH+1];
	time_t 	now;	/* time now */

	sprintf(header,"Device name: %-8s                                                  dio v%s", dev, version);
	waddstr(win, header);
	wmove(win, 1,0);
	sprintf(colhead,"                                                               | rq |     | wq |");
	waddstr(win, colhead);
	wmove(win, 2,0);
	sprintf(colhead,"    br    |    bw    |  rkb/s |  wkb/s | %%r:%%w | ro | wo |  rq | %%b |  wq | %%b |");
	waddstr(win, colhead);
	wrefresh(win);

	if(ofile != NULL){
		fprintf(ofile,"# Device name: %-8s\n", dev);

		now = time(NULL);
		fprintf(ofile,"# Started: %s\n", ctime(&now));
		fprintf(ofile,"    br    ,    bw    ,  rkb/s ,  wkb/s , %%r,%%w , ro , wo ,  rq ,rq%%b,  wq ,wq%%b\n");
	}

}


void updatescreen(WINDOW *win, WINDOW* tot, kstat_ctl_t *kc, kstat_t *ksp, int ival)
{
	kstat_io_t	kio;
	kstat_io_t	kioold;
	char		line[WIDTH];
	int			i;

	kstat_read(kc, ksp, &kio);
	sleep(ival);
	kioold = kio;

	for(i = 0; i<SCRLHEIGHT-1; i++){
		kstat_read(kc, ksp, &kio);
		calcstats(line, kio, kioold, ival);
		wmove(win,i,0);
		waddstr(win,line);
		wrefresh(win);
		printtots(tot);
		sleep(ival);
		check_input();
		kioold = kio;
	}
	
	while(1){
		kstat_read(kc, ksp, &kio);
		calcstats(line, kio, kioold, ival);
		wmove(win,SCRLHEIGHT-1,0);
		waddstr(win,line);
		scroll(win);
		wrefresh(win);
		printtots(tot);
		sleep(ival);
		check_input();
		kioold = kio;
	}
}

void calcstats(char* line, kstat_io_t	new, kstat_io_t old, int ival)
{
	u_longlong_t br;	/* bytes read */
	u_longlong_t bw;	/* bytes written */
	u_longlong_t total;	/* sum of reads and write */
	float brs;	/* bytes read per second */
	float bws;	/* bytes written per second */
	int rr;		/* read ratio */
	int wr;		/* read ratio */
	float	wq;	/* average wait queue length*/
	float	rq;	/* average run queue length*/
	float deltawtime;
	float deltawlentime;
	float deltartime;
	float deltarlentime;
	float rqbusy; /* % of time disk busy (runq occupied) */
	float wqbusy; /* % of time waitq occupied */
	uint_t reads; /* no. of read ops */
	uint_t writes; /* no. of write ops */
	
	br = new.nread - old.nread;
	bw = new.nwritten - old.nwritten;
	reads = new.reads - old.reads;
	writes = new.writes - old.writes;
	brs = (float)(br / ival / 1024.0);
	bws = (float)(bw / ival / 1024.0);
	total = br + bw;
	if(total > 0){
		rr = (int)((float)br / (float)total * 100.0);
		wr = (int)((float)bw / (float)total * 100.0);
	}
	else{
		rr = 0.0F;
		wr = 0.0F;
	}

	totr+=br;
	totw+=bw;
	if(brs > maxbrs)
		maxbrs = brs;
	if(bws > maxbws)
		maxbws = bws;

	deltawtime = (new.wtime - old.wtime) / 1000000000.0; 
	wqbusy = deltawtime / ival * 100.0;

	deltawlentime = (new.wlentime - old.wlentime) / 1000000000.0; 
	if(deltawtime > 0){
		wq = deltawlentime / ival;
	}
	else{
		wq = 0.0;
	}

	deltartime = (new.rtime - old.rtime) / 1000000000.0; 
	rqbusy = deltartime / ival * 100.0;

	deltarlentime = (new.rlentime - old.rlentime) / 1000000000.0; 
	if(deltartime > 0){
		/* maybe should be divided by ival ? */
		rq = deltarlentime / ival;
	}
	else{
		rq = 0.0;
	}

	sprintf(line,"%10lld %10lld %8.1f %8.1f %3d:%-3d %4d %4d %5.2f  %3.0f %5.2f  %3.0f", br, bw, brs, bws, rr, wr, reads, writes, rq, rqbusy, wq, wqbusy);

	if(ofile != NULL){
		fprintf(ofile, "%10lld,%10lld,%8.1f,%8.1f,%3d,%-3d,%4d,%4d,%5.2f, %3.0f,%5.2f, %3.0f\n", br, bw, brs, bws, rr, wr, reads, writes, rq, rqbusy, wq, wqbusy);
		fflush(ofile);
	}
}

void	printtots(WINDOW *win)
{
	char line[WIDTH];
	
	sprintf(line, "Max read rate:  %8.1f KB/s     Total bytes read:    %10.2f MB", maxbrs, (float)((totr) / 1048576.0));
	wmove(win,0,0);
	waddstr(win,line);
	sprintf(line, "Max write rate: %8.1f KB/s     Total bytes written: %10.2f MB", maxbws, (float)((totw) / 1048576.0));
	wmove(win,1,0);
	waddstr(win,line);
	wmove(win,2,0);
	wrefresh(win);
}

void listio(){
	kstat_ctl_t	*kcl;
	kstat_t		*kspl;
	char mapping[MAXNAMELEN];

	endwin();
	kcl = kstat_open();

	/* get device mapping */
	devdskmap();

	printf("Device Name        Class             Mapping\n\n");
	for(kspl= kcl->kc_chain; kspl!= NULL; kspl = kspl->ks_next){
		if(kspl->ks_type == KSTAT_TYPE_IO){
			if(!strcmp(kspl->ks_class,"partition")){
				getpartnmap(mapping, kspl->ks_name, kspl->ks_module);
				printf(" %-12s  ->  %-12s  ->  %s\n", 
						kspl->ks_name, kspl->ks_class, mapping);
			}
			else if(!strcmp(kspl->ks_class,"disk")){
				/* pass in the module here and ensure match 
				   with start of name */
				finddskmap(mapping, kspl->ks_name, kspl->ks_module);
				if(strcmp(mapping,"")){
					printf(" %-12s  ->  %-12s  ->  %s\n", 
						kspl->ks_name, kspl->ks_class, mapping);
				}
				else{
					printf(" %-12s  ->  %-12s\n", 
								kspl->ks_name, kspl->ks_class);
				}
			}
			else {
				printf(" %-12s  ->  %-12s\n", kspl->ks_name, kspl->ks_class);
			}
		}
		strcpy(mapping, "");
	}
	exit(0);
}

void usage(){

	printf("Usage: dio [-o <output file>] <device name> <time interval>\n");
	printf("       dio -l\n");
}

void bomb(int exitv, char *msg, int showusage){
	
	endwin();
	printf("%s",msg);
	if(showusage == 1){
		usage();
	}
	exit(exitv);
}

void check_input(){
	int c = 0;

	while(c != -1){
		c = getch();
		if(c == 4 || c == 'q'){
			/* you hit ctrl-d or q*/
			close_end(0);
		}
	}
}

void close_end(int signum) {
	time_t 	end;	/* time now */

	endwin();
	if(ofile != NULL){
		end = time(NULL);
		fprintf(ofile,"\n# Finished: %s", ctime(&end));
		fflush(ofile);
		fclose(ofile);
	}

	exit(0);
}
