/*
    dio v1.4
    Copyright 2001, 2002 Donald McIntosh
    Email: donmcintosh@hotmail.com
    Source code for dio: device I/O analysis tool

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* map sdX disks to cXtXdXsX notation */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <kstat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mkdev.h>
#include <sys/param.h>
#include <unistd.h>
#include "diskmap.h"

void mapdisk(int i, minor_t minor_num, struct dirent *dirp, char* driver);
void getdriver(char* retdriver, char* fullpath);
void findpmap(char* map, int minor, char* driver);

void devdskmap()
{
	DIR *dp;
	struct dirent *dirp;
	struct stat statbuf;
	char fullpath[MAXPATHLEN];
	char driver[MAXPATHLEN];
	minor_t minor_num;
	int i = 0, j = 0;

	if((dp=opendir("/dev/dsk"))==NULL){
		printf("Could not open directory. Exiting. \n");
		exit(-1);
	}

	while((dirp=readdir(dp))!=NULL){
		sprintf(fullpath, "/dev/dsk/%s", dirp->d_name);
		if(stat(fullpath, &statbuf) == 0){
			/* get minor number */
			minor_num = minor(statbuf.st_rdev);
		 	/* get driver */
			getdriver(driver, fullpath);

			/* assign to next space in partition map */
			partitionmap[i].minor = minor_num;
			strcpy(partitionmap[i].driver, driver);
			strcpy(partitionmap[i].name, dirp->d_name);

			/* if s0, then try to map disk too */
			if(strstr(dirp->d_name,"s0") != NULL){
				mapdisk(j, minor_num, dirp, driver);
				j++;
			}
		}
		else{
			printf("Could not stat directory %s. Exiting. \n", dirp->d_name);
			exit(-1);
		}
		i++;
	}

	totalpartn = i;
	totaldsk = j;
}

void getdriver(char* retdriver, char* fullpath)
{
	char *buffer;
	char link[MAXPATHLEN];
	char tail[MAXPATHLEN];
	char driver[MAXNAMELEN];
	int i = 0, j = 0;

	/* grab the driver name from the fullpath */

	readlink(fullpath, link, MAXPATHLEN);
	buffer = strtok(link,"/");
	while((buffer = strtok(NULL,"/")) != NULL){
		/* go through all tokens until get to end*/
		strcpy(tail, buffer);
	}

	/* now buffer contains end of fullpath */
	for(i=0; i<strlen(tail); i++){
		if(tail[i] >= 'a' && tail[i] <= 'z'){
			/* must be a letter */
			driver[j] = tail[i];
			j++;
		}
		else if (tail[i] == '@'){
			/* reached end of driver name, so term driver */
			driver[j] = NULL;
			break;
		}
	}

	strcpy(retdriver, driver);
}

void mapdisk(int i, minor_t minor_num, struct dirent *dirp, char* driver)
{
	char *buffer;
	char pdname[MAXNAMELEN];
	int diskid;

	/* diskid is the sdX number, which is always the minor num /8 */
	diskid = minor_num / 8;

	strcpy(pdname,dirp->d_name);
	/* get rid of s0 at the end */
	buffer = strtok(pdname,"s");

	dskmap[i].id = diskid;
	strcpy(dskmap[i].driver, driver);
	strcpy(dskmap[i].name, buffer);

}

void getpartnmap(char* map, char* diskname, char* driver)
{
	char sdid[40];
	int slice = 0;
	int disk_minor;
	int i = 0, j = 0;

	/* parse disk name */
	for(i=0; i<strlen(diskname); i++){
		if(diskname[i] >= '0' && diskname[i] <= '9'){
			/* must be a number */
			sdid[j] = diskname[i];
			j++;
		}
		else if (j > 0 && diskname[i] >= 'a' && diskname[i] <= 'h'){
			/* terminate sdid */
			sdid[j] = NULL;
			/* must have finished sd id, so this must be the slice */
			switch(diskname[i]){
				case 'a':
					slice = 0;
					break;
				case 'b':
					slice = 1;
					break;
				case 'c':
					slice = 2;
					break;
				case 'd':
					slice = 3;
					break;
				case 'e':
					slice = 4;
					break;
				case 'f':
					slice = 5;
					break;
				case 'g':
					slice = 6;
					break;
				case 'h':
					slice = 7;
					break;
				default:
					break;
			}
		}
		else {
			/* must be driver name still */
		}
	} 

	disk_minor = (atoi(sdid) * 8) + slice; 
	findpmap(map, disk_minor, driver);
}

void findpmap(char* map, int minor, char* driver)
{
	int i;

	for(i=0; i<totalpartn; i++){
		/* if driver and minor number match  return name*/
		if( (!strcmp(partitionmap[i].driver, driver))
				&& (minor == partitionmap[i].minor) ){
			strcpy(map, partitionmap[i].name);
			break;
		}
	}
}

void finddskmap(char* map, char* diskname, char* driver)
{
	int i = 0, j = 0;
	char sdid[40];
	int id;

	for(i=0; i<strlen(diskname); i++){
		if(diskname[i] >= '0' && diskname[i] <= '9'){
			sdid[j] = diskname[i];
			j++;
		}
	}
	/* terminate */
	sdid[j] = NULL;
	id = atoi(sdid); 

	for(i=0; i<totaldsk; i++){
		/* if driver and id match return name*/
		if( (!strcmp(dskmap[i].driver, driver))
				&& (id == dskmap[i].id) ){
			strcpy(map, dskmap[i].name);
			break;
		}
	}
}
